import marimo

__generated_with = "0.9.27"
app = marimo.App()


@app.cell
def __(mo):
    mo.md(r"""# Sample covariance""")
    return


@app.cell
def __():
    import cvxpy as cp
    import numpy as np

    from cvx.portfolio.min_risk import minrisk_problem
    from cvx.random import rand_cov
    from cvx.risk.sample import SampleCovariance

    _n = 6
    return SampleCovariance, cp, minrisk_problem, np, rand_cov


@app.cell
def __(SampleCovariance, cp, minrisk_problem):
    def problem(n):
        weights = cp.Variable(n)
        _riskmodel = SampleCovariance(num=n)
        _problem = minrisk_problem(_riskmodel, weights)

        return _problem, _riskmodel

    return (problem,)


@app.cell
def __(np, problem, rand_cov):
    _n = 50
    _a = rand_cov(_n - 2)
    _problem, _riskmodel = problem(_n)
    for _i in range(100):
        _riskmodel.update(cov=_a, lower_assets=np.zeros(48), upper_assets=np.ones(48))
        _problem.solve()
    return


@app.cell
def __(np, problem, rand_cov):
    _n = 50
    _a = rand_cov(_n - 2)
    for _i in range(100):
        _problem, _riskmodel = problem(_a.shape[0])
        _riskmodel.update(cov=_a, lower_assets=np.zeros(48), upper_assets=np.ones(48))
        _problem.solve()
    return


@app.cell
def __():
    import marimo as mo

    return (mo,)


if __name__ == "__main__":
    app.run()



import marimo

__generated_with = "0.9.27"
app = marimo.App(width="medium")


@app.cell
def __(__file__):
    from pathlib import Path

    import cvxpy as cp
    import numpy as np
    import pandas as pd

    from cvx.portfolio.min_risk import minrisk_problem
    from cvx.risk.sample import SampleCovariance
    from cvx.simulator import Builder

    pd.options.plotting.backend = "plotly"

    path = Path(__file__).parent
    return (
        Builder,
        Path,
        SampleCovariance,
        cp,
        minrisk_problem,
        np,
        path,
        pd,
    )


@app.cell
def __(path, pd):
    # Load some historic stock prices
    prices = pd.read_csv(
        path / "data" / "stock_prices.csv", index_col=0, parse_dates=True, header=0
    )

    # Estimate a series of historic covariance matrices
    returns = prices.pct_change().dropna(axis=0, how="all")
    return prices, returns


@app.cell
def __(
    Builder,
    SampleCovariance,
    cp,
    minrisk_problem,
    np,
    prices,
    returns,
):
    cov = returns.ewm(com=60, min_periods=100).cov().dropna(axis=0, how="all")
    start = cov.index[0][0]

    # Establish a risk model
    _risk_model = SampleCovariance(num=20)

    # Perform the backtest
    _builder = Builder(prices=prices.truncate(before=start), initial_aum=1e6)

    _w = cp.Variable(len(_builder.prices.columns))
    _problem = minrisk_problem(_risk_model, _w)

    for _t, _state in _builder:
        _risk_model.update(
            cov=cov.loc[_t[-1]].values,
            lower_assets=np.zeros(20),
            upper_assets=np.ones(20),
        )

        # don't reconstruct the problem in every iteration!
        _problem.solve()

        _builder.weights = _w.value
        _builder.aum = _state.aum

    _portfolio = _builder.build()

    _portfolio.nav.plot()
    return cov, start


@app.cell
def __(Builder, cp, minrisk_problem, np, prices, returns, start):
    from cvx.risk.cvar import CVar

    _risk_model = CVar(alpha=0.80, n=40, m=20)

    # Perform the backtest
    _builder = Builder(prices=prices.truncate(before=start), initial_aum=1e6)

    _w = cp.Variable(len(_builder.prices.columns))
    _problem = minrisk_problem(_risk_model, _w)

    for _t, _state in _builder:
        _risk_model.update(
            returns=returns.truncate(after=_t[-1]).tail(40).values,
            lower_assets=np.zeros(20),
            upper_assets=np.ones(20),
        )

        # don't reconstruct the problem in every iteration!
        _problem.solve()

        _builder.weights = _w.value
        _builder.aum = _state.aum

    _portfolio = _builder.build()
    _portfolio.nav.plot()
    return (CVar,)


if __name__ == "__main__":
    app.run()



from __future__ import annotations

import numpy as np
import pandas as pd
import pytest

from cvx.risk.linalg import pca


@pytest.fixture()
def returns(resource_dir):
    prices = pd.read_csv(
        resource_dir / "stock_prices.csv", index_col=0, header=0, parse_dates=True
    )
    return prices.pct_change().fillna(0.0)


def test_pca(returns):
    xxx = pca(returns)

    assert np.allclose(
        xxx.explained_variance,
        np.array(
            [
                0.33383825,
                0.19039608,
                0.11567561,
                0.07965253,
                0.06379108,
                0.04580062,
                0.03461307,
                0.02205145,
                0.01876345,
                0.01757195,
            ]
        ),
    )
    # assert False



import marimo

__generated_with = "0.9.27"
app = marimo.App(width="medium")


@app.cell
def __():
    from pathlib import Path

    import cvxpy as cvx
    import numpy as np
    import pandas as pd

    from cvx.portfolio.min_risk import minrisk_problem
    from cvx.risk.factor import FactorModel
    from cvx.risk.linalg import pca

    path = Path(__file__).parent
    return FactorModel, cvx, minrisk_problem, np, pca, path, pd


@app.cell
def __(path, pd):
    # Load some historic stock prices
    prices = pd.read_csv(
        path / "data" / "stock_prices.csv", index_col=0, parse_dates=True, header=0
    )

    # Estimate a series of historic covariance matrices
    returns = prices.pct_change().dropna(axis=0, how="all")
    return prices, returns


@app.cell
def __(pca, returns):
    factors = pca(returns=returns, n_components=10)
    return (factors,)


@app.cell
def __(FactorModel, factors, np, returns):
    model = FactorModel(assets=len(returns.columns), k=10)

    # update the model parameters
    model.update(
        cov=factors.cov,
        exposure=factors.exposure.values,
        idiosyncratic_risk=factors.idiosyncratic.std().values,
        lower_assets=np.zeros(20),
        upper_assets=np.ones(20),
        lower_factors=-0.1 * np.ones(10),
        upper_factors=0.1 * np.ones(10),
    )

    # test the risk model with uniform weights
    weights = 0.05 * np.ones(20)
    model.estimate(weights).value
    return model, weights


@app.cell
def __(cvx, minrisk_problem, model, np, pd, prices):
    w = cvx.Variable(20)
    y = cvx.Variable(10)

    problem = minrisk_problem(model, w, y=y)
    problem.solve()

    print(pd.Series(data=w.value, index=prices.columns))
    print(model.estimate(w, y=y).value)

    # check the solution
    assert np.isclose(w.value.sum(), 1.0)
    assert np.all(w.value > -0.01)
    print(y.value)
    return problem, w, y


if __name__ == "__main__":
    app.run()



#    Copyright 2023 Stanford University Convex Optimization Group
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
"""Bounds"""
from __future__ import annotations

from dataclasses import dataclass

import cvxpy as cp
import numpy as np

from .model import Model


@dataclass
class Bounds(Model):
    m: int = 0
    name: str = ""

    def estimate(self, weights, **kwargs):
        """No estimation for bounds"""
        raise NotImplementedError("No estimation for bounds")

    def _f(self, str):
        return f"{str}_{self.name}"

    def __post_init__(self):
        self.parameter[self._f("lower")] = cp.Parameter(
            shape=self.m,
            name="lower bound",
            value=np.zeros(self.m),
        )
        self.parameter[self._f("upper")] = cp.Parameter(
            shape=self.m,
            name="upper bound",
            value=np.ones(self.m),
        )

    def update(self, **kwargs):
        # lower = kwargs.get("lower", np.zeros(self.m))
        lower = kwargs[self._f("lower")]
        self.parameter[self._f("lower")].value = np.zeros(self.m)
        self.parameter[self._f("lower")].value[: len(lower)] = lower

        upper = kwargs[self._f("upper")]  # .get("upper", np.ones(self.m))
        self.parameter[self._f("upper")].value = np.zeros(self.m)
        self.parameter[self._f("upper")].value[
            : len(upper)
        ] = upper  # kwargs.get("upper", np.ones(m))

    def constraints(self, weights, **kwargs):
        return [
            weights >= self.parameter[self._f("lower")],
            weights <= self.parameter[self._f("upper")],
        ]



from __future__ import annotations

import numpy as np

from cvx.random import rand_cov
from cvx.risk.linalg import cholesky


def test_cholesky():
    a = rand_cov(10)
    u = cholesky(a)
    # test numpy arrays are equivalent
    assert np.allclose(a, np.transpose(u) @ u)



#    Copyright 2023 Stanford University Convex Optimization Group
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
"""Factor risk model
"""
from __future__ import annotations

from dataclasses import dataclass

import cvxpy as cvx
import numpy as np

from ..bounds import Bounds
from ..linalg import cholesky
from ..model import Model


@dataclass
class FactorModel(Model):
    """Factor risk model"""

    assets: int = 0
    k: int = 0

    def __post_init__(self):
        self.parameter["exposure"] = cvx.Parameter(
            shape=(self.k, self.assets),
            name="exposure",
            value=np.zeros((self.k, self.assets)),
        )

        self.parameter["idiosyncratic_risk"] = cvx.Parameter(
            shape=self.assets, name="idiosyncratic risk", value=np.zeros(self.assets)
        )

        self.parameter["chol"] = cvx.Parameter(
            shape=(self.k, self.k),
            name="cholesky of covariance",
            value=np.zeros((self.k, self.k)),
        )

        self.bounds_assets = Bounds(m=self.assets, name="assets")
        self.bounds_factors = Bounds(m=self.k, name="factors")

    def estimate(self, weights, **kwargs):
        """
        Compute the total variance
        """
        var_residual = cvx.norm2(
            cvx.multiply(self.parameter["idiosyncratic_risk"], weights)
        )

        y = kwargs.get("y", self.parameter["exposure"] @ weights)

        return cvx.norm2(
            cvx.vstack([cvx.norm2(self.parameter["chol"] @ y), var_residual])
        )

    def update(self, **kwargs):
        exposure = kwargs["exposure"]
        k, assets = exposure.shape

        self.parameter["exposure"].value[:k, :assets] = kwargs["exposure"]
        self.parameter["idiosyncratic_risk"].value[:assets] = kwargs[
            "idiosyncratic_risk"
        ]
        self.parameter["chol"].value[:k, :k] = cholesky(kwargs["cov"])
        self.bounds_assets.update(**kwargs)
        self.bounds_factors.update(**kwargs)

    def constraints(self, weights, **kwargs):
        y = kwargs.get("y", self.parameter["exposure"] @ weights)

        return (
            self.bounds_assets.constraints(weights)
            + self.bounds_factors.constraints(y)
            + [y == self.parameter["exposure"] @ weights]
        )



#    Copyright 2023 Stanford University Convex Optimization Group
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
"""Abstract risk model
"""
from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass, field

import cvxpy as cp


@dataclass
class Model(ABC):
    """Abstract risk model"""

    parameter: dict[str, cp.Parameter] = field(default_factory=dict)

    @abstractmethod
    def estimate(self, weights, **kwargs):
        """
        Estimate the variance given the portfolio weights
        """

    @abstractmethod
    def update(self, **kwargs):
        """
        Update the data in the risk model
        """

    @abstractmethod
    def constraints(self, weights, **kwargs):
        """
        Return the constraints for the risk model
        """



from __future__ import annotations

import cvxpy as cp
import numpy as np
import pytest

from cvx.risk.bounds import Bounds


def test_raise_not_implemented():
    weights = cp.Variable(3)
    bounds = Bounds(m=3, name="assets")

    with pytest.raises(NotImplementedError):
        bounds.estimate(weights)


def test_constraints():
    weights = cp.Variable(3)
    bounds = Bounds(m=3, name="assets")
    bounds.update(
        lower_assets=np.array([0.1, 0.2]), upper_assets=np.array([0.3, 0.4, 0.5])
    )

    assert bounds.parameter["lower_assets"].value == pytest.approx(
        np.array([0.1, 0.2, 0])
    )
    assert bounds.parameter["upper_assets"].value == pytest.approx(
        np.array([0.3, 0.4, 0.5])
    )

    assert len(bounds.constraints(weights)) == 2



# Import necessary libraries
from __future__ import annotations

import cvxpy as cp
import numpy as np
import pytest

from cvx.portfolio.min_risk import minrisk_problem
from cvx.risk.cvar import CVar


def test_estimate_risk():
    """Test the estimate() method"""
    model = CVar(alpha=0.95, n=50, m=14)

    np.random.seed(42)

    # define the problem
    weights = cp.Variable(14)
    prob = minrisk_problem(model, weights)
    assert prob.is_dpp()

    model.update(
        returns=np.random.randn(50, 10),
        lower_assets=np.zeros(10),
        upper_assets=np.ones(10),
    )
    prob.solve()
    assert prob.value == pytest.approx(0.5058720677762698)

    # it's enough to only update the R value...
    model.update(
        returns=np.random.randn(50, 10),
        lower_assets=np.zeros(10),
        upper_assets=np.ones(10),
    )
    prob.solve()
    assert prob.value == pytest.approx(0.43559171295408616)



#    Copyright 2023 Stanford University Convex Optimization Group
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
"""Risk models based on the sample covariance matrix
"""
from __future__ import annotations

from dataclasses import dataclass

import cvxpy as cvx
import numpy as np

from ..bounds import Bounds
from ..linalg import cholesky
from ..model import Model


@dataclass
class SampleCovariance(Model):
    """Risk model based on the Cholesky decomposition of the sample cov matrix"""

    num: int = 0

    def __post_init__(self):
        self.parameter["chol"] = cvx.Parameter(
            shape=(self.num, self.num),
            name="cholesky of covariance",
            value=np.zeros((self.num, self.num)),
        )
        self.bounds = Bounds(m=self.num, name="assets")

    def estimate(self, weights, **kwargs):
        """Estimate the risk by computing the Cholesky decomposition of self.cov"""
        return cvx.norm2(self.parameter["chol"] @ weights)

    def update(self, **kwargs):
        cov = kwargs["cov"]
        n = cov.shape[0]

        self.parameter["chol"].value[:n, :n] = cholesky(cov)
        self.bounds.update(**kwargs)

    def constraints(self, weights):
        return self.bounds.constraints(weights)



#    Copyright 2023 Stanford University Convex Optimization Group
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
from __future__ import annotations

import numpy as np


def valid(matrix):
    """
    Construct the valid subset of matrix (correlation) matrix matrix
    :param matrix: n x n matrix

    :return: Tuple of matrix boolean vector indicating if row/column
    is valid and the valid subset of the matrix
    """
    # make sure matrix  is quadratic
    if matrix.shape[0] != matrix.shape[1]:
        raise AssertionError

    v = np.isfinite(np.diag(matrix))
    return v, matrix[:, v][v]



import marimo

__generated_with = "0.9.27"
app = marimo.App()


@app.cell
def __(mo):
    mo.md(r"""# Large problem with 1000 assets and 100 factors""")
    return


@app.cell
def __():
    import cvxpy as cvx
    import numpy as np
    import pandas as pd
    from util.random import random_assets, random_beta, random_factors, random_noise

    from cvx.portfolio.min_risk import minrisk_problem
    from cvx.risk.factor import FactorModel

    return (
        FactorModel,
        cvx,
        minrisk_problem,
        np,
        pd,
        random_assets,
        random_beta,
        random_factors,
        random_noise,
    )


@app.cell
def __(random_factors):
    T = 2000
    factors = random_factors(T=T, N=100, const_factor=False)
    return T, factors


@app.cell
def __(factors, random_assets, random_beta):
    beta = random_beta(assets=random_assets(1000), factors=factors)
    return (beta,)


@app.cell
def __(beta, factors, random_noise):
    ret = factors @ beta + 0.01 * random_noise(factors @ beta)
    return (ret,)


@app.cell
def __(FactorModel, ret):
    triangle = FactorModel(assets=len(ret.columns), k=100)
    return (triangle,)


@app.cell
def __(beta, cvx, factors, minrisk_problem, np, pd, ret, triangle):
    w = cvx.Variable(1000)
    y = cvx.Variable(100)
    _problem = minrisk_problem(triangle, w, y=y)
    triangle.update(
        exposure=beta.values,
        cov=factors.cov().values,
        idiosyncratic_risk=pd.DataFrame(
            data=ret - factors @ beta, index=ret.index, columns=ret.columns
        )
        .std()
        .values,
        lower_assets=np.zeros(1000),
        upper_assets=np.ones(1000),
        lower_factors=-0.1 * np.ones(100),
        upper_factors=0.1 * np.ones(100),
    )
    return w, y


@app.cell
def __(beta, factors, minrisk_problem, np, pd, ret, triangle, w, y):
    for i in range(1):
        _problem = minrisk_problem(triangle, w, y=y)
        triangle.update(
            exposure=beta.values,
            cov=factors.cov().values,
            idiosyncratic_risk=pd.DataFrame(
                data=ret - factors @ beta, index=ret.index, columns=ret.columns
            )
            .std()
            .values,
            lower_assets=np.zeros(1000),
            upper_assets=np.ones(1000),
            lower_factors=-0.1 * np.ones(100),
            upper_factors=0.1 * np.ones(100),
        )
        _problem.solve(ignore_dpp=True)
    return (i,)


@app.cell
def __():
    import marimo as mo

    return (mo,)


if __name__ == "__main__":
    app.run()



from __future__ import annotations

import cvxpy as cp
import numpy as np

from cvx.portfolio.min_risk import minrisk_problem
from cvx.risk.sample import SampleCovariance


def test_sample():
    riskmodel = SampleCovariance(num=2)
    riskmodel.update(
        cov=np.array([[1.0, 0.5], [0.5, 2.0]]),
        lower_assets=np.zeros(2),
        upper_assets=np.ones(2),
    )
    vola = riskmodel.estimate(np.array([1.0, 1.0])).value
    np.testing.assert_almost_equal(vola, 2.0)


def test_sample_large():
    riskmodel = SampleCovariance(num=4)
    riskmodel.update(
        cov=np.array([[1.0, 0.5], [0.5, 2.0]]),
        lower_assets=np.zeros(2),
        upper_assets=np.ones(2),
    )
    vola = riskmodel.estimate(np.array([1.0, 1.0, 0.0, 0.0])).value
    np.testing.assert_almost_equal(vola, 2.0)


def test_min_variance():
    weights = cp.Variable(4)
    riskmodel = SampleCovariance(num=4)
    problem = minrisk_problem(riskmodel, weights)
    assert problem.is_dpp()

    riskmodel.update(
        cov=np.array([[1.0, 0.5], [0.5, 2.0]]),
        lower_assets=np.zeros(2),
        upper_assets=np.ones(2),
    )
    problem.solve()
    np.testing.assert_almost_equal(
        weights.value, np.array([0.75, 0.25, 0.0, 0.0]), decimal=5
    )

    # It's enough to only update the value for the cholesky decomposition
    riskmodel.update(
        cov=np.array([[1.0, 0.5], [0.5, 4.0]]),
        lower_assets=np.zeros(2),
        upper_assets=np.ones(2),
    )
    problem.solve()
    np.testing.assert_almost_equal(
        weights.value, np.array([0.875, 0.125, 0.0, 0.0]), decimal=5
    )



#    Copyright 2023 Stanford University Convex Optimization Group
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
from __future__ import annotations

from dataclasses import dataclass

import cvxpy as cvx
import numpy as np

from ..bounds import Bounds
from ..model import Model


@dataclass
class CVar(Model):
    """Conditional value at risk model"""

    alpha: float = 0.95
    n: int = 0
    m: int = 0

    def __post_init__(self):
        self.k = int(self.n * (1 - self.alpha))
        self.parameter["R"] = cvx.Parameter(
            shape=(self.n, self.m), name="returns", value=np.zeros((self.n, self.m))
        )
        self.bounds = Bounds(m=self.m, name="assets")

    def estimate(self, weights, **kwargs):
        """Estimate the risk by computing the Cholesky decomposition of self.cov"""
        # R is a matrix of returns, n is the number of rows in R
        # n = self.R.shape[0]
        # k is the number of returns in the left tail
        # k = int(n * (1 - self.alpha))
        # average value of the k elements in the left tail
        return -cvx.sum_smallest(self.parameter["R"] @ weights, k=self.k) / self.k

    def update(self, **kwargs):
        ret = kwargs["returns"]
        m = ret.shape[1]

        self.parameter["R"].value[:, :m] = kwargs["returns"]
        self.bounds.update(**kwargs)

    def constraints(self, weights, **kwargs):
        return self.bounds.constraints(weights)



from __future__ import annotations

project = "cvxrisk"

master_doc = "index"

source_suffix = {
    ".rst": "restructuredtext",
    ".md": "markdown",
}

extensions = [
    "sphinx.ext.viewcode",
    "sphinx.ext.mathjax",
    "sphinx.ext.napoleon",
    "autoapi.extension",
]


source_dir = "cvx"
autoapi_type = "python"
autoapi_dirs = [f"../../{source_dir}"]
autoapi_options = [
    "members",
    "undoc-members",
    "show-inheritance",
    "private-members",
    "show-module-summary",
    "special-members",
    "imported-members",
]

napoleon_google_docstring = True
napoleon_numpy_docstring = True

# HTML -----------------------------------------------------------------
# html_favicon = "_static/logo.jpeg"
# html_logo = "_static/logo.jpeg"
html_title = "cvxrisk Documentation"
html_show_sourcelink = True
html_css_files = [
    "custom.css",
]
# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
# html_theme = 'alabaster'
html_theme = "sphinx_rtd_theme"
# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]



from __future__ import annotations

import numpy as np
import pytest

from cvx.risk.linalg import valid


def test_valid():
    a = np.array([[np.nan, np.nan], [np.nan, 4]])
    v, mat = valid(a)

    assert np.allclose(mat, np.array([[4]]))
    assert np.allclose(v, np.array([False, True]))


def test_invalid():
    a = np.zeros((3, 2))
    with pytest.raises(AssertionError):
        valid(a)



from __future__ import annotations

import uuid

import numpy as np
import pandas as pd


def random_weights(assets):
    """
    Construct a vector of non-negative random weights. Their sum shall be 1
    """
    # Get some random weights
    weights = pd.Series(index=assets, data=np.random.rand(len(assets)))
    return weights / weights.sum()


def random_factors(T, N=2, const_factor=True):
    """
    Construct N random factor time series for T timestamps
    """
    factors = pd.DataFrame(
        index=range(1, T + 1),
        columns=[f"F{i}" for i in range(N)],
        data=np.random.randn(T, N),
    )
    # add the constant factor
    if const_factor:
        factors["const"] = 1
    return factors


def random_beta(assets, factors):
    """
    Construct a random exposure matrix
    """
    data = np.random.randn(factors.shape[1], len(assets))
    return pd.DataFrame(columns=assets, index=factors.columns, data=data)


def random_noise(frame):
    """
    Construct a frame of random noise with exactly the same dimensions as the input frame
    """
    return pd.DataFrame(
        columns=frame.columns,
        index=frame.index,
        data=np.random.randn(frame.shape[0], frame.shape[1]),
    )


def random_assets(num):
    """
    Construct a vector of random assets
    """
    return [str(uuid.uuid4())[:7] for _ in range(num)]



from __future__ import annotations

import cvxpy as cp
import numpy as np
import pandas as pd
import pytest

from cvx.portfolio.min_risk import minrisk_problem
from cvx.random import rand_cov
from cvx.risk.factor import FactorModel
from cvx.risk.linalg import pca as principal_components


@pytest.fixture()
def returns(resource_dir):
    prices = pd.read_csv(
        resource_dir / "stock_prices.csv", index_col=0, header=0, parse_dates=True
    )
    return prices.pct_change().fillna(0.0)


def test_timeseries_model(returns):
    # Here we compute the factors and regress the returns on them
    factors = principal_components(returns=returns, n_components=10)

    model = FactorModel(assets=25, k=10)

    model.update(
        cov=factors.cov.values,
        exposure=factors.exposure.values,
        idiosyncratic_risk=factors.idiosyncratic.std().values,
        lower_assets=np.zeros(20),
        upper_assets=np.ones(20),
        lower_factors=np.zeros(10),
        upper_factors=np.ones(10),
    )

    w = np.zeros(25)
    w[:20] = 0.05

    vola = model.estimate(w).value
    np.testing.assert_almost_equal(vola, 0.00923407730537884)


def test_minvar(returns):
    weights = cp.Variable(20)
    y = cp.Variable(10)

    model = FactorModel(assets=20, k=10)

    problem = minrisk_problem(model, weights, y=y)

    assert problem.is_dpp()


def test_estimate_risk():
    """Test the estimate() method"""
    model = FactorModel(assets=25, k=12)

    np.random.seed(42)

    # define the problem
    weights = cp.Variable(25)
    y = cp.Variable(12)

    prob = minrisk_problem(model, weights, y=y)
    assert prob.is_dpp()

    model.update(
        cov=rand_cov(10),
        exposure=np.random.randn(10, 20),
        idiosyncratic_risk=np.random.randn(20),
        lower_assets=np.zeros(20),
        upper_assets=np.ones(20),
        lower_factors=np.zeros(10),
        upper_factors=np.ones(10),
    )
    prob.solve()
    assert prob.value == pytest.approx(0.14138117837204583)
    assert np.array(weights.value[20:]) == pytest.approx(np.zeros(5), abs=1e-6)

    model.update(
        cov=rand_cov(10),
        exposure=np.random.randn(10, 20),
        idiosyncratic_risk=np.random.randn(20),
        lower_assets=np.zeros(20),
        upper_assets=np.ones(20),
        lower_factors=-0.1 * np.ones(10),
        upper_factors=0.1 * np.ones(10),
    )
    prob.solve()
    assert prob.value == pytest.approx(0.5454593844618784)
    assert np.array(weights.value[20:]) == pytest.approx(np.zeros(5), abs=1e-6)

    # test that the exposure is correct, e.g. the factor weights match the exposure * asset weights
    assert model.parameter["exposure"].value @ weights.value == pytest.approx(
        y.value, abs=1e-6
    )

    # test all entries of y are smaller than 0.1
    assert np.all([y.value <= 0.1 + 1e-6])
    # test all entries of y are larger than -0.1
    assert np.all([y.value >= -(0.1 + 1e-6)])



#    Copyright 2023 Stanford University Convex Optimization Group
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
"""PCA analysis
"""
from __future__ import annotations

from collections import namedtuple

import numpy as np
import pandas as pd
from sklearn.decomposition import PCA as sklearnPCA

PCA = namedtuple(
    "PCA",
    ["explained_variance", "factors", "exposure", "cov", "systematic", "idiosyncratic"],
)


def pca(returns, n_components=10):
    """
    Compute the first n principal components for a return matrix

    Args:
        returns: DataFrame of prices
        n_components: Number of components
    """

    # USING SKLEARN. Let's look at the first n components
    sklearn_pca = sklearnPCA(n_components=n_components)
    sklearn_pca.fit_transform(returns)

    exposure = sklearn_pca.components_
    factors = returns @ np.transpose(exposure)

    return PCA(
        explained_variance=sklearn_pca.explained_variance_ratio_,
        factors=factors,
        exposure=pd.DataFrame(data=exposure, columns=returns.columns),
        cov=factors.cov(),
        systematic=pd.DataFrame(
            data=factors.values @ exposure, index=returns.index, columns=returns.columns
        ),
        idiosyncratic=pd.DataFrame(
            data=returns.values - factors.values @ exposure,
            index=returns.index,
            columns=returns.columns,
        ),
    )
